import java.util.Vector;

import loci.formats.ImageReader;
import loci.tests.testng.TestTools;

public class ParallelTileTest {

  private Vector<String> hashes = new Vector<String>();

  public void run(String id) throws Exception {
    long t0 = System.currentTimeMillis();
    getHashes(id);
    long t1 = System.currentTimeMillis();
    checkParallelHashes(id);
    long t2 = System.currentTimeMillis();
    System.out.println("sequential read took " + (t1 - t0) + " ms");
    System.out.println("parallel read took " + (t2 - t1) + " ms");
  }

  private void getHashes(String id) throws Exception {
    ImageReader reader = new ImageReader();
    reader.setId(id);
    int optimalWidth = reader.getOptimalTileWidth();
    int optimalHeight = reader.getOptimalTileHeight();

    for (int y=0; y<reader.getSizeY(); y+=optimalHeight) {
      for (int x=0; x<reader.getSizeX(); x+=optimalWidth) {
        int width = (int) Math.min(optimalWidth, reader.getSizeX() - x);
        int height = (int) Math.min(optimalHeight, reader.getSizeY() - y);

        byte[] buf = reader.openBytes(0, x, y, width, height);
        hashes.add(TestTools.md5(buf));
      }
    }
    reader.close();
  }

  private void checkParallelHashes(String id) throws Exception {
    int nTiles = hashes.size();
    TileRunner[] runners = new TileRunner[nTiles];
    for (int i=0; i<nTiles; i++) {
      runners[i] = new TileRunner(id, i, hashes.get(i));
      runners[i].start();
    }
    for (int i=0; i<nTiles; i++) {
      runners[i].join();
      if (!runners[i].valid()) {
        System.out.println("*** failed on tile #" + i);
      }
    }
  }

  public static void main(String[] args) throws Exception {
    ParallelTileTest test = new ParallelTileTest();
    test.run(args[0]);
  }

  class TileRunner extends Thread {
    private ImageReader r;
    private int x = -1, y = -1, width, height;
    private String hash, validHash;

    public TileRunner(String id, int tile, String hash) {
      try {
        r = new ImageReader();
        r.setId(id);
        int optimalWidth = r.getOptimalTileWidth();
        int optimalHeight = r.getOptimalTileHeight();
        int nextTile = 0;
        for (int yy=0; yy<r.getSizeY(); yy+=optimalHeight) {
          for (int xx=0; xx<r.getSizeX(); xx+=optimalWidth) {
            if (nextTile == tile) {
              x = xx;
              y = yy;
              width = (int) Math.min(optimalWidth, r.getSizeX() - xx);
              height = (int) Math.min(optimalHeight, r.getSizeY() - yy);

              break;
            }
            else nextTile++;
          }
          if (nextTile == tile && x >=0 && y >=0) {
            break;
          }
        }
      }
      catch (Exception e) { }
      this.validHash = hash;
    }

    public void run() {
      try {
        byte[] buf = r.openBytes(0, x, y, width, height);
        hash = TestTools.md5(buf);
        r.close();
      }
      catch (Exception e) { }
    }

    public boolean valid() {
      return validHash.equals(hash);
    }

  }

}
